///////////////////////////////////////////////////////////////////////////////
//
//	app.cpp
//
///////////////////////////////////////////////////////////////////////////////
#include	<stdio.h>
#include	<string.h>
#include	<stdlib.h>
#include	<time.h>
#include	<sys/stat.h>
#include	"std.h"
#include	"mpmf20.h"

// output
#define		INFOSTR					printf
#define		ERRORSTR				INFOSTR

// platform dependencies
#if defined(_WIN32)
	// MS VC++ v5.0 for Win9x
	#include	<stdlib.h>
	#define		SIZE_MAXPATH			_MAX_PATH
	#define		DELETEARRAY				delete[]

#elif defined(__linux__)
	// linux g++
	#include	<unistd.h>
	#include	<values.h>
	#if defined(PATH_MAX)
		#define		SIZE_MAXPATH			PATH_MAX
	#else
		#define		SIZE_MAXPATH			256
	#endif
	#define		DELETEARRAY				delete[]

#else
	// not supported
	#error ! ! compiler/platform not supported ! !
#endif

// default port base
#if defined(__alpha)
	#define		PORT_BASE_DEFAULT		0x3bc
#else
	#define		PORT_BASE_DEFAULT		0x378
#endif

// return code
#define		CLEANUP_RETURN( ret )		\
{ 										\
	if ( pDevice ) 						\
	{ 									\
		delete pDevice; 				\
		pDevice = NULL; 				\
	} 									\
	if ( pszFilePlaylistTemp ) 			\
	{ 									\
		unlink( pszFilePlaylistTemp );	\
		pszFilePlaylistTemp = NULL; 	\
	} 									\
	return ret; 						\
}

///////////////////////////////////////////////////////////////////////////////
// return file size
static long GetFileSize( char* pszPathFile )
{
	long lReturn = 0;

	FILE* fpFile = fopen( pszPathFile, "rb" );
	if ( fpFile )
	{
		struct stat sStat;
		if ( !stat(pszPathFile, &sStat) )
			lReturn = sStat.st_size;

		fclose( fpFile );
	}

	return lReturn;
}

///////////////////////////////////////////////////////////////////////////////
// progress callback
static BOOL ProgressCallback( int iPos, int iCount )
{
	INFOSTR( "remaining %-5hd\r", iCount-iPos );
	fflush( stdout );
	return TRUE;
}

///////////////////////////////////////////////////////////////////////////////
// display directory
static void DisplayDirectory( CMPMF20& cMPMF20, BOOL bVerbose )
{
	CDir& cDir = cMPMF20.GetDirectory();
	int iCountEntry = cDir.m_iCountEntry;

	INFOSTR( "\n" );
	INFOSTR( "   entry count: %d\n", iCountEntry );
	INFOSTR( "  total memory: %ld KB\n", ((long)cDir.m_iCountBlockAvailable * cMPMF20.GetBlockSize()) / 1024 );
	INFOSTR( "   used memory: %ld KB\n", ((long)cDir.m_iCountBlockUsed * cMPMF20.GetBlockSize()) / 1024 );
	INFOSTR( " unused memory: %ld KB\n", ((long)cDir.m_iCountBlockRemaining * cMPMF20.GetBlockSize()) / 1024 );
	INFOSTR( "flash ram type: %s\n", cMPMF20.GetUseExternalFlashStatus() ? "external" : "internal" );

	if ( bVerbose )
		INFOSTR( "    bad blocks: %ld\n", (long)cDir.m_iCountBlockBad );

	if ( iCountEntry )
	{
		CDirEntry* pDirEntry = cDir.m_acDirEntry;
		if ( iCountEntry > CMPMF20_MAX_DIRENTRY )
			iCountEntry = CMPMF20_MAX_DIRENTRY;

		INFOSTR( "\n" );
		INFOSTR( "No  Size      Title\n" );
		INFOSTR( "-----------------------------------------------------------------------------\n" );

		for( int iA=0; iA<iCountEntry; ++iA, ++pDirEntry )
			INFOSTR( "%02u  %-8ld  %-40.40s\n", iA+1, pDirEntry->m_lSize, pDirEntry->m_szName );
	}
}

///////////////////////////////////////////////////////////////////////////////
// process playlist file
static BOOL ProcessPlaylist( CMPMF20& cMPMF20, char* pszFile, BOOL bVerbose )
{
	char szBuf[ SIZE_MAXPATH ];

	// open playlist for read
	FILE* fpFile = fopen( pszFile, "r" );
	if ( !fpFile )
	{
		ERRORSTR( "unable to open '%s' for read\n", pszFile );
		return FALSE;
	}

	// check if device can accommodate all files in playlist
	CDir& cDir = cMPMF20.GetDirectory();
	long lSizeAvailable = (long)cDir.m_iCountBlockAvailable * cMPMF20.GetBlockSize();
	long lSizeCurrent = (long)cDir.m_iCountBlockUsed * cMPMF20.GetBlockSize();
	int iCountEntryCurrent = cDir.m_iCountEntry;
	while( fgets(szBuf, sizeof(szBuf), fpFile) )
	{
		// strip 'new line' char
		szBuf[ strlen(szBuf)-1 ] = 0;

		// get file size
		long lSize = GetFileSize( szBuf );
		if ( !lSize )
		{
			ERRORSTR( "unable to open '%s' for read\n", szBuf );
			fclose( fpFile );
			return FALSE;
		}

		// check space
		lSizeCurrent += lSize;
		if ( lSizeCurrent > lSizeAvailable )
		{
			ERRORSTR( "entry '%s' exceed's memory capacity of device\n", szBuf );
			fclose( fpFile );
			return FALSE;
		}

		// check enough entries
		++iCountEntryCurrent;
		if ( iCountEntryCurrent > CMPMF20_MAX_DIRENTRY )
		{
			ERRORSTR( "entry '%s' exceed's maximum directory entry count for device\n", szBuf );
			fclose( fpFile );
			return FALSE;
		}
	}

	// rewind playlist
	rewind( fpFile );

	// process entries
	while( fgets(szBuf, sizeof(szBuf), fpFile) )
	{
		// strip 'new line' char
		szBuf[ strlen(szBuf)-1 ] = 0;

		// upload
		if ( bVerbose )
			INFOSTR( "uploading %s\n", szBuf );
		if ( !cMPMF20.TxFile(szBuf, bVerbose ? ProgressCallback : NULL) )
		{
			ERRORSTR( "tx file failed, %s\n", cMPMF20.GetErrorStr() );
			fclose( fpFile );
			return FALSE;
		}
	}

	// close playlist file
	fclose( fpFile );

	return TRUE;
}

///////////////////////////////////////////////////////////////////////////////
// change playlist order
static BOOL ChangePlaylistOrder( CMPMF20& cMPMF20, char* pszPlaylistOrder )
{
	UINT auiPosNew[ CMPMF20_MAX_DIRENTRY ];

	UINT uiPosNew = 0;
	char* pszEntry;
	do
	{
		pszEntry = strtok( uiPosNew ? NULL : pszPlaylistOrder, " " );
		if ( pszEntry )
			auiPosNew[ uiPosNew++ ] = atoi( pszEntry ) - 1;
	} while( pszEntry && (uiPosNew < CMPMF20_MAX_DIRENTRY) );

	if ( !cMPMF20.SetFileOrder(auiPosNew, uiPosNew) )
	{
		ERRORSTR( "change in playlist order failed, %s\n", cMPMF20.GetErrorStr() );
		return FALSE;
	}

	return TRUE;
}

///////////////////////////////////////////////////////////////////////////////
// display help
static void Help( void )
{
	INFOSTR( "\nmpmf20 utility v%d.%02d - Ashpool Systems (c) 1999\n", CMPMF20_ID_VERSION/100, CMPMF20_ID_VERSION%100 );
	INFOSTR( "-----------------------------------------------\n" );
	INFOSTR( "command line switches available :-\n" );
	INFOSTR( "	-d  display directory\n" );
	INFOSTR( "	-i  initialize with check for bad blocks\n" );
	INFOSTR( "	-x  perform operations on external flash ram\n" );
	INFOSTR( "	-u  specify file(s) to upload\n" );
	INFOSTR( "	-g  specify file to download\n" );
	INFOSTR( "	-f  specify text based playlist file which contains files to be upload\n" );
	INFOSTR( "	-z  specify file to delete\n" );
	INFOSTR( "	-za delete all files\n" );
	INFOSTR( "	-o  specify new playlist order in quotes\n" );
	INFOSTR( "	-p  specify parallel port base IO address, default=0x%x\n", PORT_BASE_DEFAULT );
	INFOSTR( "	-v  enable verbose mode\n" );

	INFOSTR( "\nexamples...\n" );
	INFOSTR( "	; display directory using parallel port at 0x278\n" );
	INFOSTR( "		mpmf20 -p 0x278 -d\n" );
	INFOSTR( "	; initialize and upload files\n" );
	INFOSTR( "		mpmf20 -i -u *.mp3\n" );
	INFOSTR( "	; delete existing files and upload playlist onto external flash ram\n" );
	INFOSTR( "		mpmf20 -za -f playlist.txt -x\n" );
	INFOSTR( "	; initialize, upload files in playlist and then display directory\n" );
	INFOSTR( "		mpmf20 -d -i -f playlist.txt\n" );
	INFOSTR( "	; download file then delete it and finally display directory\n" );
	INFOSTR( "		mpmf20 -d -g mp3Files/song.mp3 -z song.mp3\n" );
	INFOSTR( "	; reverse playlist order and then display directory\n" );
	INFOSTR( "		mpmf20 -o \"5 4 3 2 1\" -d\n" );
}

///////////////////////////////////////////////////////////////////////////////
int main( int iCountArg, char* paszArg[] )
{
	// default settings
	BOOL bDisplayDir = FALSE;
	BOOL bInit = FALSE;
	BOOL bUseExternalFlash = FALSE;
	BOOL bDeleteAll = FALSE;
	BOOL bVerbose = FALSE;
	CMPMF20* pDevice = NULL;
	char* pszFileDelete = NULL;
	char* pszFileDownload = NULL;
	char* pszFilePlaylist = NULL;
	char* pszFilePlaylistTemp = NULL;
	char* pszPlaylistOrder = NULL;
	int iPortBase = PORT_BASE_DEFAULT;

	// process command line args
	if ( iCountArg < 2 )
	{
		Help();
		CLEANUP_RETURN( FALSE );
	}
	for( int iA=1; iA<iCountArg; ++iA )
	{
		// check for display directory request
		if ( !strcmp(paszArg[iA], "-d") )
			bDisplayDir = TRUE;
		// check for init with mark bad block request
		else if ( !strcmp(paszArg[iA], "-i") )
			bInit = TRUE;
		// check for delete all file request
		else if ( !strcmp(paszArg[iA], "-za") )
			bDeleteAll = TRUE;
		// check for request to use external flash memory
		else if ( !strcmp(paszArg[iA], "-x") )
			bUseExternalFlash = TRUE;
		// check for verbose mode request
		else if ( !strcmp(paszArg[iA], "-v") )
			bVerbose = TRUE;
		// check for delete file request
		else if ( !strcmp(paszArg[iA], "-z") )
		{
			if ( (iA+1) < iCountArg )
			{
				++iA;
				pszFileDelete = paszArg[iA];
			}
		}
		// check for upload file request
		else if ( !strcmp(paszArg[iA], "-u") )
		{
			if ( (iA+1) < iCountArg )
			{
				++iA;

				// append all command line upload requests to temp playlist
				if ( !pszFilePlaylistTemp )
					pszFilePlaylistTemp = tmpnam( NULL );

				if ( pszFilePlaylistTemp )
				{
					FILE* fpFile = fopen( pszFilePlaylistTemp, "a" );
					if ( !fpFile )
					{
						ERRORSTR( "unable to open '%s' for amend\n", pszFilePlaylistTemp );
						CLEANUP_RETURN( FALSE );
					}
					while( iA < iCountArg )
					{
						fprintf( fpFile, "%s\n", paszArg[iA] );
						if ( (iA+1) < iCountArg )
						{
							if ( paszArg[iA+1][0] == '-' )
								break;
						}
						++iA;
					}
					fclose( fpFile );
				}
			}
		}
		// check for download file request
		else if ( !strcmp(paszArg[iA], "-g") )
		{
			if ( (iA+1) < iCountArg )
			{
				++iA;
				pszFileDownload = paszArg[iA];
			}
		}
		// check for playlist file request
		else if ( !strcmp(paszArg[iA], "-f") )
		{
			if ( (iA+1) < iCountArg )
			{
				++iA;
				pszFilePlaylist = paszArg[iA];
			}
		}
		// check for playlist order request
		else if ( !strcmp(paszArg[iA], "-o") )
		{
			if ( (iA+1) < iCountArg )
			{
				++iA;
				pszPlaylistOrder = paszArg[iA];
			}
		}
		// check for port base request
		else if ( !strcmp(paszArg[iA], "-p") )
		{
			if ( (iA+1) < iCountArg )
			{
				++iA;
				sscanf( paszArg[iA], "%x", &iPortBase );
			}
		}
		// else help
		else
		{
			Help();
			CLEANUP_RETURN( FALSE );
		}
	}

	// setup
	pDevice = new CMPMF20;
	if ( !pDevice )
	{
		ERRORSTR( "not enough memory\n" );
		CLEANUP_RETURN( FALSE );
	}
	CMPMF20& cMPMF20 = *pDevice;
	if ( !cMPMF20.Set(iPortBase) )
	{
		ERRORSTR( "%s\n", cMPMF20.GetErrorStr() );
		CLEANUP_RETURN( FALSE );
	}

	// select internal or external flash memory
	if ( bVerbose )
		INFOSTR( "performing operations on %s flash ram\n", bUseExternalFlash ? "external" : "internal" );
	cMPMF20.UseExternalFlash( bUseExternalFlash );

	// check device present
	if ( bVerbose )
		INFOSTR( "check device present\n" );
	if ( !cMPMF20.CheckPresent() )
	{
		ERRORSTR( "%s\n", cMPMF20.GetErrorStr() );
		ERRORSTR( "most common problems are ...\n" );
		ERRORSTR( "    Incorrect port address.\n" );
		ERRORSTR( "    Parallel port is not set to ECP or EPP mode.\n" );
		ERRORSTR( "    Device is not on.\n" );
		ERRORSTR( "    Battery power is too low.\n" );
		CLEANUP_RETURN( FALSE );
	}

	// update directory flag
	BOOL bUpdateDirectory = FALSE;

	// get current directory
	if ( bVerbose )
		INFOSTR( "downloading directory\n" );
	if ( !cMPMF20.RxDirectory() )
	{
		if ( cMPMF20.GetErrorID() == CMPMF20_ERROR_CORRUPT )
			ERRORSTR( "%s\n", cMPMF20.GetErrorStr() );
		else
		{
			ERRORSTR( "rx directory failed, %s\n", cMPMF20.GetErrorStr() );
			CLEANUP_RETURN( FALSE );
		}
	}

	// if init request
	if ( bInit )
	{
		if ( bVerbose )
			INFOSTR( "initializing with bad block check enabled\n" );

		if ( !cMPMF20.Initialize( bVerbose ? ProgressCallback : NULL) )
		{
			ERRORSTR( "Initialize failed, %s\n", cMPMF20.GetErrorStr() );
			CLEANUP_RETURN( FALSE );
		}

		bUpdateDirectory = TRUE;
	}

	// if playlist request
	if ( pszFilePlaylist )
	{
		if ( !ProcessPlaylist(cMPMF20, pszFilePlaylist, bVerbose) )
			CLEANUP_RETURN( FALSE );

		bUpdateDirectory = TRUE;
	}

	// if temp playlist request
	if ( pszFilePlaylistTemp )
	{
		if ( !ProcessPlaylist(cMPMF20, pszFilePlaylistTemp, bVerbose) )
			CLEANUP_RETURN( FALSE );

		bUpdateDirectory = TRUE;
	}

	// if download request
	if ( pszFileDownload )
	{
		if ( bVerbose )
			INFOSTR( "downloading %s\n", pszFileDownload );

		if ( !cMPMF20.RxFile(pszFileDownload, bVerbose ? ProgressCallback : NULL) )
		{
			ERRORSTR( "rx file failed, %s\n", cMPMF20.GetErrorStr() );
			CLEANUP_RETURN( FALSE );
		}
	}

	// if delete request
	if ( pszFileDelete )
	{
		if ( bVerbose )
			INFOSTR( "deleting %s\n", pszFileDelete );

		if ( !cMPMF20.RemoveFile(pszFileDelete) )
		{
			ERRORSTR( "delete file failed, %s\n", cMPMF20.GetErrorStr() );
			CLEANUP_RETURN( FALSE );
		}

		bUpdateDirectory = TRUE;
	}

	// if delete all request
	if ( bDeleteAll )
	{
		if ( bVerbose )
			INFOSTR( "deleting all files\n" );

		if ( !cMPMF20.RemoveAllFiles() )
		{
			ERRORSTR( "delete all files failed, %s\n", cMPMF20.GetErrorStr() );
			CLEANUP_RETURN( FALSE );
		}

		bUpdateDirectory = TRUE;
	}

	// if playlist order request
	if ( pszPlaylistOrder )
	{
		if ( bVerbose )
			INFOSTR( "changing playlist order\n" );

		if ( !ChangePlaylistOrder(cMPMF20, pszPlaylistOrder) )
			CLEANUP_RETURN( FALSE );

		bUpdateDirectory = TRUE;
	}

	// if directory update required
	if ( bUpdateDirectory )
	{
		if ( bVerbose )
			INFOSTR( "updating directory\n" );

		if ( !cMPMF20.TxDirectory() )
		{
			ERRORSTR( "tx directory failed, %s\n", cMPMF20.GetErrorStr() );
			CLEANUP_RETURN( FALSE );
		}
	}

	// if display directory request
	if ( bDisplayDir )
		DisplayDirectory( cMPMF20, bVerbose );

	// restore
	CLEANUP_RETURN( TRUE );
}

