///////////////////////////////////////////////////////////////////////////////
//
//	mpmf20.cpp
//
///////////////////////////////////////////////////////////////////////////////
#include	<stdio.h>
#include	<string.h>
#include	<stdarg.h>
#include	<time.h>
#include	<errno.h>
#include	<sys/stat.h>
#include	"std.h"
#include	"binary.h"
#include	"mpmf20.h"

// platform dependencies
#if defined(_WIN32)
	// MS VC++ v5.0 for Win9x
	#include	<conio.h>
	#define		OUTPORT( p, v )			_outp( (p), (v) )
	#define		INPORT( p )				_inp( (p) )
	#define		CLOCK_SECOND			1000
	#define		DELETEARRAY				delete[]
	#define		STRICMP					stricmp

#elif defined(__linux__)
	// linux g++
	#include	<unistd.h>
	#if defined(__alpha)
		#include	<sys/io.h>
	#else
		#include	<sys/perm.h>
	#endif
	#include	<asm/io.h>
	#define		OUTPORT(p,v)			outb( (v), (p) )
	#define		INPORT(p)				inb( (p) )
	#define		CLOCK_SECOND			1000000
	#define		DELETEARRAY				delete[]
	#define		STRICMP					strcasecmp

#else
	// not supported
	#error ! ! compiler/platform not supported ! !
#endif

// port offset constants
#define		OFFSET_PORT_DATA		0
#define		OFFSET_PORT_STATUS		1
#define		OFFSET_PORT_CONTROL		2

// base block
#define		BASE_BLOCK_INTERNAL		0x0000
#define		BASE_BLOCK_EXTERNAL		0x3000

// size, count of block, page and page tag
#define		SIZE_TAG				16
#define		SIZE_PAGE				512
#define		COUNT_PAGEDEFAULT		16
#define		COUNT_PAGEMAX			32
#define		SIZE_BLOCKMAX			(SIZE_PAGE * COUNT_PAGEMAX)
#define		SIZE_TAGALLMAX			(SIZE_TAG * COUNT_PAGEMAX)

// Block type
#define		ID_BLOCK_BAD			0x00
#define		ID_BLOCK_VALIDMP3		0x6d
#define		ID_BLOCK_NOTVALIDMP3	0x65
#define		ID_BLOCK_NOTAVAILABLE	0xcc
#define		ID_BLOCK_FREE			0xff

// retry delay
#define		TIME_RETRYDELAY			(CLOCK_SECOND / 4)
// max tx/rx block retry
#define		MAX_RETRY				8

// max duration of init
#define		TIME_INIT				40

// delay's
#define		IODELAY(c)				{ for( int _iA=0; _iA<(c); ++_iA ) INPORT(m_iPortStatus); }
#define		DELAY(t)				{ long _lTime=clock()+(t); while( _lTime > clock() ) ; }

// command
#define		DATAOUT(v1, v2, v3)	{ OUTPORT(m_iPortControl, (v2)); OUTPORT(m_iPortData, (v1) ); OUTPORT(m_iPortControl, (v3) ); IODELAY(20); }
#define		WAITSTATUS( v )		{ if ( !WaitStatus(v) ) return FALSE; }
#define		DATAOUTWAITACK( v ) { DATAOUT( (v), B_00001100, B_00000100); IODELAY(100); }

// byte order conversion
#define		SHORTCONVERT( v )		( \
										(((USHORT)(v) & 0x00ff) << 8) | \
										(((USHORT)(v) & 0xff00) >> 8) \
									)

#define		UNSIGNED32CONVERT( v )	( \
										(((UNSIGNED32)(v) & 0x000000ff) << 24) | \
										(((UNSIGNED32)(v) & 0x0000ff00) <<  8) | \
										(((UNSIGNED32)(v) & 0x00ff0000) >>  8) | \
										(((UNSIGNED32)(v) & 0xff000000) >> 24) \
									)

///////////////////////////////////////////////////////////////////////////////
// hex dump
void HexDump( UCHAR* pauc, int iSize )
{
	int iA, iB;

	UCHAR* pucS = pauc;

	int iXXX = iSize;

	if ( iSize % 16 )
		iSize = ((iSize / 16)+1) * 16;

	for( iA=0; iA<(iSize/16); ++iA, pucS+=16 )
	{
		printf( "%08lx  ", iA*16 );
		for( iB=0; iB<16; ++iB )
		{
			if ( (iA * 16 + iB) < iXXX )
			{
				int iC = *(pucS+iB);
				printf( "%02x ", iC );
			}
			else
				printf( "   " );

			if ( (iB%8) == 7 )
				printf( " " );
		}
		for( iB=0; iB<16; ++iB )
		{
			if ( (iA * 16 + iB) < iXXX )
			{
				int iC = *(pucS+iB);
				printf( "%c", iC>=32 && iC<255 ? iC : '.' );
			}
			else
				printf( " " );

			if ( (iB%8) == 7 )
				printf( " " );
		}
		printf( "\n" );
		if ( (iA%8) == 7 )
			printf( "\n" );
	}

	printf( "\n" );
}

// create checksum
static UINT CreateChecksum( void* pv, int iSize )
{
	if ( !pv )
		return 0;

	UCHAR ucValue = 0;
	UCHAR* pauc = (UCHAR*)pv;
	for( int iA=0; iA<iSize; ++iA )
		ucValue += (*pauc++);

	return ucValue;
}

// return pointer to file
static char* GetFile( char* pszPathFile )
{
	int iLength = strlen( pszPathFile );
	if ( !iLength )
		return pszPathFile;

	char* pc = pszPathFile + iLength - 1;
	while( *pc != '\\' && *pc != '/' && *pc != ':' )
	{
		if ( pc == pszPathFile )
			return pc;
		--pc;
	}
	++pc;

	return pc;
}

// return pointer to file extention
static char* GetFileExtension( char* szFile )
{
	int iLength = strlen( szFile );
	if ( !iLength )
		return szFile;

	// start from last char of filename
	char* pc = szFile + iLength - 1;
	while( *pc != '.' )
	{
		// if no extension in filename, return null string
		if ( pc == szFile )
			return szFile + iLength;
		--pc;
	}
	++pc;
	return pc;
}

///////////////////////////////////////////////////////////////////////////////
// end of block tag
void CEndOfBlockTag::Convert( void )
{
	m_usOffsetBlock = SHORTCONVERT( m_usOffsetBlock );
	m_usPosBlockPrev = SHORTCONVERT( m_usPosBlockPrev );
	m_usPosBlock = SHORTCONVERT( m_usPosBlock );
	m_usPosBlockNext = SHORTCONVERT( m_usPosBlockNext );
	m_usSize = SHORTCONVERT( m_usSize );
	m_usNotUsed = SHORTCONVERT( m_usNotUsed );
}

void CEndOfBlockTag::Reset( void )
{
	memset( this, 0, sizeof(CEndOfBlockTag) );
}

///////////////////////////////////////////////////////////////////////////////
// set/unset, constructors and destructors
void CMPMF20::Unset( void )
{
}

BOOL CMPMF20::Set( int iPortBase )
{
	// cleanup any previous Set()
	Unset();

	// determine ports
	m_iPortBase = iPortBase;
	m_iPortData = m_iPortBase + OFFSET_PORT_DATA;
	m_iPortStatus = m_iPortBase + OFFSET_PORT_STATUS;
	m_iPortControl = m_iPortBase + OFFSET_PORT_CONTROL;

	// init error
	m_szError[ 0 ] = 0;
	m_iIDError = CMPMF20_ERROR_NONE;

	// if linux
	#if defined(__linux__)
		// request access to required ports
		if ( ioperm(m_iPortBase, 3, 1) )
		{
			LogError( CMPMF20_ERROR_IOPRERM, "ioperm() failed, reason '%s'", SZERROR );
			return FALSE;
		}
	#endif

	return TRUE;
}

CMPMF20::CMPMF20()
{
	// default to use internal flash
	m_bUseExternalFlash = FALSE;
	m_iPosBlockBase = BASE_BLOCK_INTERNAL;
	m_iCountPage = COUNT_PAGEDEFAULT;
	m_iSizeBlock = m_iCountPage * SIZE_PAGE;
}

CMPMF20::~CMPMF20()
{
	Unset();
}

///////////////////////////////////////////////////////////////////////////////
// log error
void CMPMF20::LogError( int iIDError, const char* pszFormat, ... )
{
	va_list ArgPtr;
	va_start( ArgPtr, pszFormat );
	vsprintf( m_szError, pszFormat, ArgPtr );
	va_end( ArgPtr );

	m_iIDError = iIDError;
}

// find first free block
int CMPMF20::FindFirstFreeBlock( void )
{
	for( int iA=m_cDir.m_iCountBlockAvailable-1; iA>=2; --iA )
	{
		if ( m_cDir.m_aucBlockType[iA] == ID_BLOCK_FREE )
			return iA;
	}

	return 0;
}

// set block type
void CMPMF20::SetBlockType( int iBlock, int iType )
{
	m_cDir.m_aucBlockType[ iBlock ] = iType;
}

// wait status
BOOL CMPMF20::WaitStatus( int iValue )
{
	int iIn;

	for( int iA=0; iA<32768; ++iA )
	{
		iIn = INPORT( m_iPortStatus ) & 0xf0;
		if ( iIn == iValue )
			return TRUE;
	}

	return FALSE;
}

// tx block retry
BOOL CMPMF20::TxBlockRetry( void* pv, void* pvTag, BOOL bTagAll, UINT uiPosBlock )
{
	int iA, iB;

	UINT uiHi = uiPosBlock / 256;
	UINT uiLo = uiPosBlock & 255;

	// issue tx command
	DATAOUTWAITACK( B_00000101 );
	DATAOUTWAITACK( B_01001010 );
	DATAOUTWAITACK( uiHi );
	DATAOUTWAITACK( uiLo );
	DATAOUTWAITACK( B_00000100 );

	// process all page's in block
	UCHAR* pauc = (UCHAR*)pv;
	UCHAR* pauc2 = (UCHAR*)pvTag;
	for( iA=0; iA<m_iCountPage; ++iA )
	{
		// start of page negotiation
		OUTPORT( m_iPortControl, B_00000000 );
		WAITSTATUS( B_10110000 );

		// process page
		for( iB=0; iB<SIZE_PAGE; ++iB )
		{
			OUTPORT( m_iPortData, *pauc++ );
			OUTPORT( m_iPortControl, B_00000010 );
			OUTPORT( m_iPortControl, B_00000000 );
		}

		// process end of page data
		if ( ((!iA) || bTagAll) && pvTag )
		{
			for( iB=0; iB<SIZE_TAG; ++iB )
			{
				OUTPORT( m_iPortData, *pauc2++ );
				OUTPORT( m_iPortControl, B_00000010 );
				OUTPORT( m_iPortControl, B_00000000 );
			}
		}
		else
		{
			for( iB=0; iB<SIZE_TAG; ++iB )
			{
				OUTPORT( m_iPortData, 0 );
				OUTPORT( m_iPortControl, B_00000010 );
				OUTPORT( m_iPortControl, B_00000000 );
			}
		}

		// end of page negotiation
		OUTPORT( m_iPortData, B_00000000 );
		OUTPORT( m_iPortControl, B_00000100 );
		if ( iA == (m_iCountPage-1) )
		{
			WAITSTATUS( B_01110000 );
		}
		else
		{
			WAITSTATUS( B_00110000 );
		}
	}

	return TRUE;
}

// tx block
BOOL CMPMF20::TxBlock( void* pv, void* pvTag, BOOL bTagAll, UINT uiPosBlock )
{
	UCHAR ucChecksumBlock = CreateChecksum( pv, m_iSizeBlock );
	UCHAR ucChecksumTag;
	if ( !bTagAll )
		ucChecksumTag = CreateChecksum( pvTag, SIZE_TAG );
	else
		ucChecksumTag = CreateChecksum( pvTag, SIZE_TAG * m_iCountPage );

	int iRetry = 0;
	while( iRetry < MAX_RETRY )
	{
		if ( TxBlockRetry(pv, pvTag, bTagAll, uiPosBlock) )
		{
			UCHAR aucBlock[ SIZE_BLOCKMAX ];
			UCHAR aucTag[ SIZE_TAGALLMAX ];

			if ( RxBlockRetry(aucBlock, aucTag, bTagAll, uiPosBlock) )
			{
				UCHAR ucChecksumBlock2 = CreateChecksum( aucBlock, m_iSizeBlock );
				UCHAR ucChecksumTag2 = 0;
				if ( pvTag )
				{
					if ( !bTagAll )
						ucChecksumTag2 = CreateChecksum( pvTag, SIZE_TAG );
					else
						ucChecksumTag2 = CreateChecksum( pvTag, SIZE_TAG * m_iCountPage );
				}

				if ( ucChecksumBlock == ucChecksumBlock2 && ucChecksumTag == ucChecksumTag2 )
					return TRUE;
			}

			//printf( "TxBlock() checksum failed, block 0x%04x retry %d\n", uiPosBlock, iRetry );
		}

		//printf( "TxBlock() failed, block 0x%04x retry %d\n", uiPosBlock, iRetry );

		DELAY( TIME_RETRYDELAY );
		++iRetry;
	}
	LogError( CMPMF20_ERROR_TXBLOCKRETRY, "too many retries for tx block" );
	return FALSE;

}

// rx block retry
BOOL CMPMF20::RxBlockRetry( void* pv, void* pvTag, BOOL bTagAll, UINT uiPosBlock )
{
	int iA, iB;

	UINT uiHi = uiPosBlock / 256;
	UINT uiLo = uiPosBlock & 255;

	// issue rx command
	DATAOUTWAITACK( B_00000101 );
	DATAOUTWAITACK( B_01000111 );
	DATAOUTWAITACK( uiHi );
	DATAOUTWAITACK( uiLo );
	DATAOUTWAITACK( B_00000100 );

	// process all page's in block
	UCHAR* pauc = (UCHAR*)pv;
	UCHAR* pauc2 = (UCHAR*)pvTag;
	for( iA=0; iA<m_iCountPage; ++iA )
	{
		// start of page negotiation
		OUTPORT( m_iPortControl, B_00000000 );
		WAITSTATUS( B_10110000 );

		// process page
		for( iB=0; iB<SIZE_PAGE; ++iB )
		{
			OUTPORT( m_iPortControl, B_00100001 );
			OUTPORT( m_iPortControl, B_00100000 );
			*pauc++ = INPORT( m_iPortData );
		}

		// process end of page data
		if ( ((!iA) || bTagAll) && pvTag )
		{
			for( iB=0; iB<SIZE_TAG; ++iB )
			{
				OUTPORT( m_iPortControl, B_00100001 );
				OUTPORT( m_iPortControl, B_00100000 );
				*pauc2++ = INPORT( m_iPortData );
			}
		}
		else
		{
			for( iB=0; iB<SIZE_TAG; ++iB )
			{
				OUTPORT( m_iPortControl, B_00100001 );
				OUTPORT( m_iPortControl, B_00100000 );
				INPORT( m_iPortData );
			}
		}

		// end of page negotiation
		OUTPORT( m_iPortControl, B_00000100 );
		OUTPORT( m_iPortData, B_00000100 );
		if ( iA == (m_iCountPage-1) )
		{
			WAITSTATUS( B_01110000 );
		}
		else
		{
			WAITSTATUS( B_00110000 );
		}
	}

	return TRUE;
}

// rx block
BOOL CMPMF20::RxBlock( void* pv, void* pvTag, BOOL bTagAll, UINT uiPosBlock )
{
	int iRetry = 0;
	while( iRetry < MAX_RETRY )
	{
		if ( RxBlockRetry(pv, pvTag, bTagAll, uiPosBlock) )
		{
			UCHAR ucChecksumBlock = CreateChecksum( pv, m_iSizeBlock );
			UCHAR ucChecksumTag;
			if ( !bTagAll )
				ucChecksumTag = CreateChecksum( pvTag, SIZE_TAG );
			else
				ucChecksumTag = CreateChecksum( pvTag, SIZE_TAG * m_iCountPage );

			if ( RxBlockRetry(pv, pvTag, bTagAll, uiPosBlock) )
			{
				UCHAR ucChecksumBlock2 = CreateChecksum( pv, m_iSizeBlock );
				UCHAR ucChecksumTag2;
				if ( !bTagAll )
					ucChecksumTag2 = CreateChecksum( pvTag, SIZE_TAG );
				else
					ucChecksumTag2 = CreateChecksum( pvTag, SIZE_TAG * m_iCountPage );

				if ( ucChecksumBlock == ucChecksumBlock2 && ucChecksumTag == ucChecksumTag2 )
					return TRUE;
			}

			//printf( "RxBlock() checksum failed, block 0x%04x retry %d\n", uiPosBlock, iRetry );
		}

		//printf( "RxBlock() failed, block 0x%04x retry %d\n", uiPosBlock, iRetry );

		DELAY( TIME_RETRYDELAY );
		++iRetry;
	}
	LogError( CMPMF20_ERROR_RXBLOCKRETRY, "too many retries for rx block" );
	return FALSE;
}

// rx page retry
BOOL CMPMF20::RxPageRetry( void* pv, void* pvTag, UINT uiPosBlock,
	UINT uiPosPage )
{
	int iA;

	UINT uiHi = uiPosBlock / 256;
	UINT uiLo = uiPosBlock & 255;

	// issue rx page
	DATAOUTWAITACK( B_00000101 );
	DATAOUTWAITACK( B_01000101 );
	DATAOUTWAITACK( uiHi );
	DATAOUTWAITACK( uiLo );
	DATAOUTWAITACK( uiPosPage );
	DATAOUTWAITACK( B_00000100 );

	// start of negotiation
	OUTPORT( m_iPortControl, B_00000000 );
	WAITSTATUS( B_10110000 );

	// process page
	UCHAR* pauc = (UCHAR*)pv;
	for( iA=0; iA<SIZE_PAGE; ++iA )
	{
		OUTPORT( m_iPortControl, B_00100001 );
		OUTPORT( m_iPortControl, B_00100000 );
		*pauc++ = INPORT( m_iPortData );
	}

	// process end of page data
	if ( pvTag )
	{
		UCHAR* pauc2 = (UCHAR*)pvTag;
		for( iA=0; iA<SIZE_TAG; ++iA )
		{
			OUTPORT( m_iPortControl, B_00100001 );
			OUTPORT( m_iPortControl, B_00100000 );
			*pauc2++ = INPORT( m_iPortData );
		}
	}
	else
	{
		for( iA=0; iA<SIZE_TAG; ++iA )
		{
			OUTPORT( m_iPortControl, B_00100001 );
			OUTPORT( m_iPortControl, B_00100000 );
			INPORT( m_iPortData );
		}
	}

	// end of page negotiation
	OUTPORT( m_iPortControl, B_00000100 );
	OUTPORT( m_iPortData, B_00000100 );
	WAITSTATUS( B_01110000 );

	return TRUE;
}

// rx page
BOOL CMPMF20::RxPage( void* pv, void* pvTag, UINT uiPosBlock, UINT uiPosPage )
{
	int iRetry = 0;
	while( iRetry < MAX_RETRY )
	{
		if ( RxPageRetry(pv, pvTag, uiPosBlock, uiPosPage) )
		{
			UCHAR ucChecksumBlock = CreateChecksum( pv, SIZE_PAGE );
			UCHAR ucChecksumTag = CreateChecksum( pvTag, SIZE_TAG );

			if ( RxPageRetry(pv, pvTag, uiPosBlock, uiPosPage) )
			{
				UCHAR ucChecksumBlock2 = CreateChecksum( pv, SIZE_PAGE );
				UCHAR ucChecksumTag2 = CreateChecksum( pvTag, SIZE_TAG );

				if ( ucChecksumBlock == ucChecksumBlock2 && ucChecksumTag == ucChecksumTag2 )
					return TRUE;

				//printf( "RxPage() checksum failed, block 0x%04x page 0x%04x retry %d\n", uiPosBlock, uiPosPage, iRetry );
			}
		}

		//printf( "RxPage() failed, block 0x%04x page 0x%04x retry %d\n", uiPosBlock, uiPosPage, iRetry );

		DELAY( TIME_RETRYDELAY );
		++iRetry;
	}
	return FALSE;
}

// mark bad blocks
BOOL CMPMF20::MarkBadBlocks( BOOL (*pfProgress)(int iPos, int iCount) )
{
	// request init with bad block check
	DATAOUTWAITACK( B_00000101 );
	DATAOUTWAITACK( m_bUseExternalFlash ? B_01100011 : B_01000011 );
	DATAOUTWAITACK( B_00000100 );

	OUTPORT( m_iPortControl, B_00000000 );
	OUTPORT( m_iPortData, B_11111111 );
	WAITSTATUS( B_11110000 );

	DATAOUT( B_00000101, B_00001100, B_00000100 );

	long lTime = time( NULL ) + TIME_INIT;
	while( lTime > time(NULL) )
	{
		if ( pfProgress )
		{
			long lTime2 = lTime - time( NULL );
			if ( lTime2 < 0 )
				lTime2 = 0;
			if ( !pfProgress(TIME_INIT-lTime2, TIME_INIT) )
			{
				LogError( CMPMF20_ERROR_INTERRUPTED, "operation interrupted" );
				return FALSE;
			}
		}

		if ( (INPORT(m_iPortStatus) & 0xf0) == B_01110000 )
			return TRUE;
	}

	return FALSE;
}

///////////////////////////////////////////////////////////////////////////////
// operations
void CMPMF20::UseExternalFlash( BOOL bUseExternalFlash )
{
	m_bUseExternalFlash = bUseExternalFlash;

	if ( !m_bUseExternalFlash )
	{
		m_iPosBlockBase = BASE_BLOCK_INTERNAL;
		m_iCountPage = COUNT_PAGEDEFAULT;
	}
	else
	{
		m_iPosBlockBase = BASE_BLOCK_EXTERNAL;
		m_iCountPage = COUNT_PAGEMAX;
	}

	m_iSizeBlock = m_iCountPage * SIZE_PAGE;
}

BOOL CMPMF20::CheckPresent( void )
{
	UCHAR aucBlock[ SIZE_PAGE ];
	UCHAR aucTag[ SIZE_TAG ];

	// check for first 512 bytes
	if ( !RxPage(aucBlock, aucTag, m_iPosBlockBase, 0) )
	{
		LogError( CMPMF20_ERROR_DEVICENOTFOUND, "device not found" );
		return FALSE;
	}

	return TRUE;
}

CDirEntry* CMPMF20::FindFile( char* pszFile )
{
	// search directory entries for matching filename
	int iCountEntry = m_cDir.m_iCountEntry;
	CDirEntry* pDirEntry = m_cDir.m_acDirEntry;
	for( int iA=0; iA<iCountEntry; ++iA, ++pDirEntry )
	{
		if ( !strcmp(pszFile, pDirEntry->m_szName) )
			return pDirEntry;
	}

	return NULL;
}

BOOL CMPMF20::Initialize( BOOL (*pfProgress)(int iPos, int iCount) )
{
	// mark bad block
	if ( !MarkBadBlocks(pfProgress) )
		return FALSE;

	// rx directory
	if ( !RxDirectory() )
		return FALSE;

	return TRUE;
}

BOOL CMPMF20::RemoveFile( char* pszFile )
{
	// get directory entry for file
	CDirEntry* pDirEntry = FindFile( pszFile );
	if ( !pDirEntry )
	{
		LogError( CMPMF20_ERROR_FILENOTFOUND, "file '%s' not present on device", pszFile );
		return FALSE;
	}

	// free FAT and blocks used
	UINT uiPosBlock = pDirEntry->m_uiPosBlock;
	while( uiPosBlock )
	{
		m_cDir.m_aucBlockType[ uiPosBlock ] = ID_BLOCK_FREE;
		UINT uiTemp = m_cDir.m_ausFAT[ uiPosBlock ];
		m_cDir.m_ausFAT[ uiPosBlock ] = 0;
		uiPosBlock = uiTemp;
	}

	// adjust directory
	--m_cDir.m_iCountEntry;
	m_cDir.m_iCountBlockUsed -= pDirEntry->m_uiCountBlock;
	m_cDir.m_iCountBlockRemaining += pDirEntry->m_uiCountBlock;

	// shuffle directory entries
	int iPosEntry = pDirEntry - m_cDir.m_acDirEntry;
	int iCount = m_cDir.m_iCountEntry - iPosEntry;
	for( int iA=0; iA<iCount; ++iA )
	{
		memcpy(
			&m_cDir.m_acDirEntry[ iPosEntry+iA ],
			&m_cDir.m_acDirEntry[ iPosEntry+iA+1 ],
			sizeof(CDirEntry)
		);
	}

	return TRUE;
}

BOOL CMPMF20::RemoveAllFiles( void )
{
	while( m_cDir.m_iCountEntry > 0 )
		RemoveFile( m_cDir.m_acDirEntry[0].m_szName );

	return TRUE;
}

BOOL CMPMF20::SetFileOrder( UINT* pauiPosOrder, UINT uiCount )
{
	// create copy of directory entries
	CDirEntry* paDirEntry = m_cDir.m_acDirEntry;
	CDirEntry* paDirEntryNew = new CDirEntry[ CMPMF20_MAX_DIRENTRY ];
	if ( !paDirEntryNew )
	{
		LogError( CMPMF20_ERROR_ALLOC, "new failed" );
		return FALSE;
	}
	memcpy( paDirEntryNew, paDirEntry, CMPMF20_MAX_DIRENTRY * sizeof(CDirEntry) );

	// current entry count
	UINT uiCountEntry = m_cDir.m_iCountEntry;

	// move entries
	for( UINT uiPosNew=0; uiPosNew<uiCount; ++uiPosNew )
	{
		UINT uiPosOld = pauiPosOrder[ uiPosNew ];

		if ( uiPosOld >= uiCountEntry || uiPosNew >= uiCountEntry )
		{
			LogError( CMPMF20_ERROR_INVALIDFILEPOSITION, "invalid file position" );
			DELETEARRAY paDirEntryNew;
			return FALSE;
		}

		memcpy( &paDirEntryNew[uiPosNew], &paDirEntry[uiPosOld], sizeof(CDirEntry) );
	}

	// update current directory entries
	memcpy( paDirEntry, paDirEntryNew, CMPMF20_MAX_DIRENTRY * sizeof(CDirEntry) );

	DELETEARRAY paDirEntryNew;

	return TRUE;
}

BOOL CMPMF20::TxDirectory( void )
{
	UCHAR aucBlock[ SIZE_BLOCKMAX ];
	UCHAR aucTag[ SIZE_TAGALLMAX ];
	CDirEntry* pDirEntry;
	UINT uiPosBlock, uiPosBlockDir;
	USHORT* pus;
	int iA, iB;

	// build dirblock
	memset( aucBlock, ID_BLOCK_FREE, m_cDir.m_iCountBlockAvailable );
	memset( aucBlock+m_cDir.m_iCountBlockAvailable, ID_BLOCK_NOTAVAILABLE, SIZE_BLOCKMAX-m_cDir.m_iCountBlockAvailable );
	for( iA=0; iA<m_cDir.m_iCountBlockAvailable; ++iA )
	{
		if ( m_cDir.m_aucBlockType[iA] == ID_BLOCK_BAD )
			aucBlock[ iA ] = ID_BLOCK_BAD;
	}
	pDirEntry = m_cDir.m_acDirEntry;
	pus = ((USHORT*)aucBlock) + 1;
	for( iA=0; iA<m_cDir.m_iCountEntry; ++iA, ++pDirEntry, ++pus )
	{
		if ( pDirEntry->m_bValidMP3 )
			*pus = 0x6d4d;
		else
			*pus = 0x6545;

		uiPosBlock = pDirEntry->m_uiPosBlock;
		while( uiPosBlock )
		{
			if ( aucBlock[uiPosBlock] == ID_BLOCK_FREE )
				aucBlock[ uiPosBlock ] = pDirEntry->m_bValidMP3 ? ID_BLOCK_VALIDMP3 : ID_BLOCK_NOTVALIDMP3;
			uiPosBlock = m_cDir.m_ausFAT[ uiPosBlock ];
		}
	}
	// set dirblock tag
	memset( aucTag, ID_BLOCK_NOTAVAILABLE, sizeof(aucTag) );
	for( iA=0; iA<(m_cDir.m_iCountBlockAvailable/SIZE_PAGE); ++iA )
	{
		UCHAR* pauc = aucTag + iA * SIZE_TAG;
		memset( pauc, 0xff, SIZE_TAG );
		*(USHORT*)pauc = SHORTCONVERT( m_cDir.m_usRamType );
	}
	// tx dirblock
	if ( !TxBlock(aucBlock, aucTag, TRUE, m_iPosBlockBase) )
		return FALSE;

	// process entries
	pDirEntry = m_cDir.m_acDirEntry;
	for( iA=0; iA<m_cDir.m_iCountEntry; ++iA, ++pDirEntry )
	{
		// clear
		memset( aucBlock, 0, sizeof(aucBlock) );

		// dir block
		uiPosBlockDir = GetDirectoryEntryPos( pDirEntry );

		// FAT
		pus = (USHORT*)aucBlock;
		*pus++ = SHORTCONVERT( m_iPosBlockBase + uiPosBlockDir );
		*pus++ = SHORTCONVERT( m_iPosBlockBase + uiPosBlockDir + 1 );
		uiPosBlock = pDirEntry->m_uiPosBlock;
		while( uiPosBlock )
		{
			*pus++ = SHORTCONVERT( m_iPosBlockBase + uiPosBlock );
			uiPosBlock = m_cDir.m_ausFAT[ uiPosBlock ];
		}
		memset( aucTag, 0, sizeof(aucTag) );
		pus = (USHORT*)aucTag;
		*(pus+0) = SHORTCONVERT( m_iPosBlockBase + uiPosBlockDir + 1 );
		*(pus+5) = SHORTCONVERT( m_iPosBlockBase + uiPosBlockDir + 1 );

		if ( !TxBlock(aucBlock, pus, FALSE, m_iPosBlockBase + uiPosBlockDir) )
			return FALSE;

		// filesize, filename
		memset( aucBlock, 0, sizeof(aucBlock) );
		memset( aucTag, 0, sizeof(aucTag) );
		CEndOfDirEntryTag* pEndOfDirEntryTag = (CEndOfDirEntryTag*)aucTag;
		if ( pDirEntry->m_bValidMP3 )
			strcpy( pEndOfDirEntryTag->m_szType, "mp3" );
		else
			strcpy( pEndOfDirEntryTag->m_szType, "ext" );
		pEndOfDirEntryTag->m_usCountBlock = SHORTCONVERT( pDirEntry->m_uiCountBlock );
		pEndOfDirEntryTag->m_usPosBlockPrev = SHORTCONVERT( m_iPosBlockBase + uiPosBlockDir );
		pEndOfDirEntryTag->m_usPosBlock = SHORTCONVERT( m_iPosBlockBase + uiPosBlockDir+1 );
		pEndOfDirEntryTag->m_usPosBlockNext = SHORTCONVERT( m_iPosBlockBase + pDirEntry->m_uiPosBlock );
		pEndOfDirEntryTag->m_usSizeBMP = SHORTCONVERT( 0x0100 );
		pEndOfDirEntryTag->m_usNotUsed = SHORTCONVERT( 0x0100 );
		pEndOfDirEntryTag->m_lSize = UNSIGNED32CONVERT( pDirEntry->m_lSize );
		STRNCPYSAFE( pEndOfDirEntryTag->m_szName, pDirEntry->m_szName, sizeof(pEndOfDirEntryTag->m_szName) );

		if ( !TxBlock(aucBlock, pEndOfDirEntryTag, TRUE, m_iPosBlockBase+uiPosBlockDir+1) )
			return FALSE;
	}

	return TRUE;
}

BOOL CMPMF20::RxDirectory( void )
{
	UCHAR aucBlock[ SIZE_BLOCKMAX ];
	UCHAR aucTag[ SIZE_TAG ];
	int iA, iB, iC;

	//printf( "---- BASE 0x%04x PAGECOUNT %d SIZEBLOCK %d\n", m_iPosBlockBase, m_iCountPage, m_iSizeBlock );

	// get directory block
	if ( !RxBlock(aucBlock, aucTag, FALSE, m_iPosBlockBase) )
		return FALSE;

	// set dir
	memset( m_cDir.m_ausFAT, 0, sizeof(m_cDir.m_ausFAT) );
	memset( m_cDir.m_aucBlockType, ID_BLOCK_FREE, sizeof(m_cDir.m_aucBlockType) );
	for( iA=0; iA<CMPMF20_MAX_BLOCK; ++iA )
	{
		if ( aucBlock[iA] == ID_BLOCK_NOTAVAILABLE )
			break;
	}
	m_cDir.m_iCountBlockAvailable = iA;
	m_cDir.m_iCountBlockUsed = 2;
	m_cDir.m_iCountBlockRemaining = 0;
	m_cDir.m_iCountBlockBad = 0;
	m_cDir.m_iCountEntry = 0;
	m_cDir.m_usRamType = SHORTCONVERT( (*(USHORT*)aucTag) );

	// entry count
	USHORT* pus = (USHORT*)aucBlock;
	CDirEntry* pDirEntry = m_cDir.m_acDirEntry;
	for( iA=0; iA<m_cDir.m_iCountBlockAvailable; ++iA )
	{
		if ( aucBlock[iA] != ID_BLOCK_FREE )
			++m_cDir.m_iCountBlockUsed;

		if ( aucBlock[iA] == ID_BLOCK_BAD )
		{
			m_cDir.m_aucBlockType[ iA ] = ID_BLOCK_BAD;
			++m_cDir.m_iCountBlockBad;
		}

		if ( !(iA & 1) )
		{
			if ( m_cDir.m_iCountEntry < CMPMF20_MAX_DIRENTRY )
			{
				if ( *pus == 0x6d4d )
				{
					pDirEntry->m_bValidMP3 = TRUE;
					++pDirEntry;
					++m_cDir.m_iCountEntry;
				}
				else if ( *pus == 0x6545 )
				{
					pDirEntry->m_bValidMP3 = FALSE;
					++pDirEntry;
					++m_cDir.m_iCountEntry;
				}
			}

			++pus;
		}
	}

	// determine remaining memory count
	m_cDir.m_iCountBlockRemaining = m_cDir.m_iCountBlockAvailable -
		m_cDir.m_iCountBlockUsed;

	// process entries
	pDirEntry = m_cDir.m_acDirEntry;
	for( iA=0; iA<m_cDir.m_iCountEntry; ++iA, ++pDirEntry )
	{
		// first two block contains FAT and filesize and filename
		UINT uiPosBlockDir = GetDirectoryEntryPos( pDirEntry );

		// process fat
		UINT uiPosBlockPrev = 0;
		UINT uiPosBlock;
		for( iB=0; iB<m_iCountPage; ++iB )
		{
			// get page
			if ( !RxPage(aucBlock, aucTag, m_iPosBlockBase+uiPosBlockDir, iB) )
				return FALSE;

			// process entries
			USHORT* pus = (USHORT*)aucBlock;
			for( iC=0; iC<(SIZE_PAGE/2); ++iC, ++pus )
			{
				// block
				uiPosBlock = SHORTCONVERT( *pus );
				UINT uiPosBlockRelative;
				if ( uiPosBlock )
				{
					uiPosBlockRelative = uiPosBlock - (UINT)m_iPosBlockBase;
					if ( uiPosBlockRelative >= CMPMF20_MAX_BLOCK )
					{
						LogError( CMPMF20_ERROR_CORRUPT, "directory corrupt, initialization recommended" );
						return FALSE;
					}
				}

				// note first audio block
				if ( iB==0 && iC==2 )
					pDirEntry->m_uiPosBlock = uiPosBlockRelative;

				if ( !uiPosBlock )
					break;

				// update FAT
				if ( uiPosBlockPrev )
					m_cDir.m_ausFAT[ uiPosBlockPrev ] = uiPosBlockRelative;

				// block type
				m_cDir.m_aucBlockType[ uiPosBlockRelative ] = ID_BLOCK_NOTVALIDMP3;

				uiPosBlockPrev = uiPosBlockRelative;
			}
			if ( iC < (SIZE_PAGE/2) )
				break;
		}

		// process all pages to get entry size and name
		for( iB=1; iB<m_iCountPage; ++iB )
		{
			// get page
			if ( !RxPage(aucBlock, aucTag, m_iPosBlockBase+uiPosBlockDir+1, iB) )
				return FALSE;

			// if first page
			if ( iB==1 )
			{
				pDirEntry->m_lSize = UNSIGNED32CONVERT( *(UNSIGNED32*)aucTag );
				pDirEntry->m_uiCountBlock = pDirEntry->m_lSize / m_iSizeBlock;
				pDirEntry->m_uiSizeMod = pDirEntry->m_lSize % m_iSizeBlock;
				if ( pDirEntry->m_uiSizeMod )
					++pDirEntry->m_uiCountBlock;
				memcpy( pDirEntry->m_szName, &aucTag[sizeof(SIGNED32)], SIZE_TAG-sizeof(SIGNED32) );
			}
			else
				memcpy( &pDirEntry->m_szName[(iB-1)*SIZE_TAG-sizeof(SIGNED32)], aucTag, SIZE_TAG );
		}
	}

	return TRUE;
}

BOOL CMPMF20::TxFile( char* pszPathFile, BOOL (*pfProgress)(int iPos, int iCount) )
{
	UCHAR aucBlock[ SIZE_BLOCKMAX ];
	int iA, iB;

	// if no enough room in directory for entry
	if ( m_cDir.m_iCountEntry >= CMPMF20_MAX_DIRENTRY )
	{
		LogError( CMPMF20_ERROR_MAXDIRENTRY, "max number of directory entries (%d) already in use", CMPMF20_MAX_DIRENTRY );
		return FALSE;
	}

	// open file for read
	FILE* fpFile = fopen( pszPathFile, "rb" );
	if ( !fpFile )
	{
		LogError( CMPMF20_ERROR_OPEN, "unable to open '%s' for read", pszPathFile );
		return FALSE;
	}

	// get fileinfo
	struct stat sStat;
	if ( stat(pszPathFile, &sStat) )
	{
		LogError( CMPMF20_ERROR_STAT, "stat() failed for '%s'", pszPathFile );
		fclose( fpFile );
		return FALSE;
	}

	// determine if enough room to store file
	int iCountBlock = (sStat.st_size / m_iSizeBlock) + 1;
	if ( (iCountBlock+2) > m_cDir.m_iCountBlockRemaining )
	{
		LogError( CMPMF20_ERROR_MEMORY, "not enough memory on device to upload '%s'", pszPathFile );
		fclose( fpFile );
		return FALSE;
	}

	// point to directory entry that will be used
	CDirEntry& cDirEntry = m_cDir.m_acDirEntry[ m_cDir.m_iCountEntry ];

	// update directory
	++m_cDir.m_iCountEntry;
	m_cDir.m_iCountBlockUsed += iCountBlock;
	m_cDir.m_iCountBlockRemaining -= iCountBlock;

	// find first free block
	int iPosBlockFree = FindFirstFreeBlock();
	if ( !iPosBlockFree )
	{
		LogError( CMPMF20_ERROR_CORRUPT, "no free blocks, initialization recommended" );
		fclose( fpFile );
		return FALSE;
	}

	// update directory entry
	memset( &cDirEntry, 0, sizeof(CDirEntry) );
	STRNCPYSAFE( cDirEntry.m_szName, GetFile(pszPathFile), sizeof(cDirEntry.m_szName) );
	cDirEntry.m_lSize = sStat.st_size;
	cDirEntry.m_uiPosBlock = iPosBlockFree;
	cDirEntry.m_uiCountBlock = iCountBlock;
	cDirEntry.m_uiSizeMod = cDirEntry.m_lSize % m_iSizeBlock;
	cDirEntry.m_bValidMP3 = (STRICMP( GetFileExtension(pszPathFile), "mp3" ) == 0);

	// default return
	BOOL bResult = FALSE;

	// process all blocks of file
	ULONG ulTxed = 0;
	UINT uiPosBlockCurrent = cDirEntry.m_uiPosBlock;
	UINT uiPosBlockPrev = GetDirectoryEntryPos( &cDirEntry );
	UINT uiPosBlockEnd = cDirEntry.m_uiCountBlock;
	UINT uiPosBlockNext;
	UINT uiPosBlock;
	for( uiPosBlock=0; uiPosBlock<uiPosBlockEnd; ++uiPosBlock )
	{
		// progress callback
		if ( pfProgress )
		{
			if ( !pfProgress(uiPosBlock, uiPosBlockEnd) )
			{
				LogError( CMPMF20_ERROR_INTERRUPTED, "operation interrupted" );
				break;
			}
		}

		// mark as block used
		SetBlockType( uiPosBlockCurrent, cDirEntry.m_bValidMP3 ? ID_BLOCK_VALIDMP3 : ID_BLOCK_NOTVALIDMP3 );

		// get next block and update FAT
		if ( uiPosBlock == (uiPosBlockEnd-1) )
		{
			uiPosBlockNext = 0xffff;
			m_cDir.m_ausFAT[ uiPosBlockCurrent ] = 0;
		}
		else
		{
			uiPosBlockNext = FindFirstFreeBlock();
			if ( uiPosBlockNext == 0 )
			{
				LogError( CMPMF20_ERROR_CORRUPT, "no free blocks, initialization recommended" );
				break;
			}
			m_cDir.m_ausFAT[ uiPosBlockCurrent ] = uiPosBlockNext;
		}

		// determine read size
		UINT uiSize;
		if ( ulTxed >= (ULONG)cDirEntry.m_lSize )
			uiSize = 0;
		else
		{
			if ( uiPosBlock == (uiPosBlockEnd-1) && cDirEntry.m_uiSizeMod )
			{
				uiSize = cDirEntry.m_uiSizeMod;
				memset( aucBlock+uiSize, 0, SIZE_BLOCKMAX-uiSize );
			}
			else
				uiSize = m_iSizeBlock;
		}

		// read block
		memset( aucBlock, 0, sizeof(aucBlock) );
		if ( uiSize )
		{
			if ( fread(aucBlock, uiSize, 1, fpFile) < 1 )
			{
				LogError( CMPMF20_ERROR_READ, "error reading from file '%s'", pszPathFile );
				break;
			}
		}

		// end block tag
		CEndOfBlockTag cEndOfBlockTag;
		if ( cDirEntry.m_bValidMP3 )
			strcpy( cEndOfBlockTag.m_szType, "mp3" );
		else
			strcpy( cEndOfBlockTag.m_szType, "ext" );
		cEndOfBlockTag.m_usOffsetBlock = uiPosBlock;
		cEndOfBlockTag.m_usPosBlockPrev = m_iPosBlockBase + uiPosBlockPrev;
		cEndOfBlockTag.m_usPosBlock = m_iPosBlockBase + uiPosBlockCurrent;
		if ( uiPosBlockNext == 0xffff )
			cEndOfBlockTag.m_usPosBlockNext = uiPosBlockNext;
		else
			cEndOfBlockTag.m_usPosBlockNext = m_iPosBlockBase + uiPosBlockNext;
		cEndOfBlockTag.m_usSize = uiSize;
		cEndOfBlockTag.m_usNotUsed = 0;

		//printf( "TX off %-4hd curr 0x%04hx prev 0x%04hx next 0x%04hx\n", cEndOfBlockTag.m_usOffsetBlock, cEndOfBlockTag.m_usPosBlock, cEndOfBlockTag.m_usPosBlockPrev, cEndOfBlockTag.m_usPosBlockNext );

		cEndOfBlockTag.Convert();

		// tx
		if ( !TxBlock(aucBlock, &cEndOfBlockTag, FALSE, m_iPosBlockBase + uiPosBlockCurrent) )
			break;

		// update current and prev block markers
		ulTxed += uiSize;
		uiPosBlockPrev = uiPosBlockCurrent;
		uiPosBlockCurrent = uiPosBlockNext;
	}

	// if transfer ok
	if ( uiPosBlock == uiPosBlockEnd )
		bResult = TRUE;

	// close file
	fclose( fpFile );

	return bResult;
}

BOOL CMPMF20::RxFile( char* pszPathFile, BOOL (*pfProgress)(int iPos, int iCount) )
{
	UCHAR aucBlock[ SIZE_BLOCKMAX ];
	int iA, iB;

	// get directory entry for file
	char* pszFile = GetFile( pszPathFile );
	CDirEntry* pDirEntry = FindFile( pszFile );
	if ( !pDirEntry )
	{
		LogError( CMPMF20_ERROR_FILENOTFOUND, "file '%s' not present on device", pszFile );
		return FALSE;
	}

	// open file for write
	FILE* fpFile = fopen( pszPathFile, "wb" );
	if ( !fpFile )
	{
		LogError( CMPMF20_ERROR_OPEN, "unable to open '%s' for write", pszPathFile );
		return FALSE;
	}

	// default return
	BOOL bResult = FALSE;

	// get all blocks
	UINT uiPosBlockCurrent = pDirEntry->m_uiPosBlock;
	UINT uiPosBlockEnd = pDirEntry->m_uiCountBlock;
	UINT uiPosBlock;
	for( uiPosBlock=0; uiPosBlock<uiPosBlockEnd; ++uiPosBlock )
	{
		// progress callback
		if ( pfProgress )
		{
			if ( !pfProgress(uiPosBlock, uiPosBlockEnd) )
			{
				LogError( CMPMF20_ERROR_INTERRUPTED, "operation interrupted" );
				break;
			}
		}

		// rx block
		if ( !RxBlock(aucBlock, NULL, FALSE, m_iPosBlockBase+uiPosBlockCurrent) )
			break;

		// determine write size
		int iSizeWrite;
		if ( uiPosBlock == (uiPosBlockEnd-1) && pDirEntry->m_uiSizeMod )
			iSizeWrite = pDirEntry->m_uiSizeMod;
		else
			iSizeWrite = m_iSizeBlock;

		// save block
		if ( fwrite(aucBlock, iSizeWrite, 1, fpFile) < 1 )
		{
			LogError( CMPMF20_ERROR_WRITE, "error writing to file '%s'", pszPathFile );
			break;
		}

		// get current block
		uiPosBlockCurrent = m_cDir.m_ausFAT[ uiPosBlockCurrent ];
	}

	// if transfer ok
	if ( uiPosBlock == uiPosBlockEnd )
		bResult = TRUE;

	// close file
	fclose( fpFile );

	return bResult;
}

