///////////////////////////////////////////////////////////////////////////////
//
//	mpmf20.h
//
///////////////////////////////////////////////////////////////////////////////

#ifndef		_MPMF20_
#define		_MPMF20_

// dependencies
#include	"std.h"

// version
#define		CMPMF20_ID_VERSION				101

// max directory entries
#define		CMPMF20_MAX_DIRENTRY			64

// max blocks
#define		CMPMF20_MAX_BLOCK				8192

// 32 bit word
#if defined(__alpha)
	typedef int SIGNED32;
	typedef unsigned int UNSIGNED32;
#else
	typedef long SIGNED32;
	typedef unsigned long UNSIGNED32;
#endif

// error id's
enum
{
	CMPMF20_ERROR_NONE = 0,
	CMPMF20_ERROR_ALLOC,
	CMPMF20_ERROR_IOPRERM,
	CMPMF20_ERROR_CREATEFILE,
	CMPMF20_ERROR_DEVICENOTFOUND,
	CMPMF20_ERROR_DRIVERVERSION,
	CMPMF20_ERROR_TXBLOCKRETRY,
	CMPMF20_ERROR_RXBLOCKRETRY,
	CMPMF20_ERROR_CORRUPT,
	CMPMF20_ERROR_FILENOTFOUND,
	CMPMF20_ERROR_INVALIDFILEPOSITION,
	CMPMF20_ERROR_MAXDIRENTRY,
	CMPMF20_ERROR_MEMORY,
	CMPMF20_ERROR_OPEN,
	CMPMF20_ERROR_STAT,
	CMPMF20_ERROR_READ,
	CMPMF20_ERROR_WRITE,
	CMPMF20_ERROR_INTERRUPTED,
};

///////////////////////////////////////////////////////////////////////////////
// enable byte align structs
#pragma pack( 1 )

// dir entry
struct CDirEntry
{
	char m_szName[ 512 ];
	long m_lSize;
	UINT m_uiPosBlock;
	UINT m_uiCountBlock;
	UINT m_uiSizeMod;
	BOOL m_bValidMP3;
};

// directory block
struct CDir
{
	USHORT m_ausFAT[ CMPMF20_MAX_BLOCK ];
	UCHAR m_aucBlockType[ CMPMF20_MAX_BLOCK ];
	CDirEntry m_acDirEntry[ CMPMF20_MAX_DIRENTRY ];
	int m_iCountBlockAvailable;
	int m_iCountBlockUsed;
	int m_iCountBlockRemaining;
	int m_iCountBlockBad;
	int m_iCountEntry;
	USHORT m_usRamType;
};

// end of block tag
struct CEndOfBlockTag
{
	char m_szType[ 4 ];
	USHORT m_usOffsetBlock;
	USHORT m_usPosBlockPrev;

	USHORT m_usPosBlock;
	USHORT m_usPosBlockNext;
	USHORT m_usSize;
	USHORT m_usNotUsed;

	void Convert( void );
	void Reset( void );
};

// end of dir entry tag
struct CEndOfDirEntryTag
{
	char m_szType[ 4 ];
	USHORT m_usCountBlock;
	USHORT m_usPosBlockPrev;
	USHORT m_usPosBlock;
	USHORT m_usPosBlockNext;
	USHORT m_usSizeBMP;
	USHORT m_usNotUsed;
	SIGNED32 m_lSize;
	char m_szName[ 256 - 20 ];
};

// disable byte align
#pragma pack()

///////////////////////////////////////////////////////////////////////////////
// mpmf20 class
class CMPMF20
{
protected:

	// directory
	CDir m_cDir;

	// error str
	char m_szError[ 128 ];

	// port constants
	int m_iPortBase;
	int m_iPortData;
	int m_iPortStatus;
	int m_iPortControl;

	// block base
	int m_iPosBlockBase;
	// page count
	int m_iCountPage;
	// block size
	int m_iSizeBlock;

	// error
	int m_iIDError;

	// internal/external flash ram flag
	BOOL m_bUseExternalFlash;

	void LogError( int iType, const char* pszFormat, ... );
	int FindFirstFreeBlock( void );
	void SetBlockType( int iBlock, int iType );
	BOOL WaitStatus( int iValue );

	BOOL TxBlockRetry( void* pv, void* pvTag, BOOL bTagAll, UINT uiPosBlock );
	BOOL TxBlock( void* pv, void* pvTag, BOOL bTagAll, UINT uiPosBlock );

	BOOL RxBlockRetry( void* pv, void* pvTag, BOOL bTagAll, UINT uiPosBlock );
	BOOL RxBlock( void* pv, void* pvTag, BOOL bTagAll, UINT uiPosBlock );

	BOOL RxPageRetry( void* pv, void* pvTag, UINT uiPosBlock,
		UINT uiPosPage );
	BOOL RxPage( void* pv, void* pvTag, UINT uiPosBlock,
		UINT uiPosPage );

	BOOL MarkBadBlocks( BOOL (*pfProgress)(int iPos, int iCount) );

public:

	// set/unset
	void Unset( void );
	BOOL Set( int iPortBase );

	// constructors/destructors
	CMPMF20();
	~CMPMF20();

	// retrieval
	CDir& GetDirectory( void ) { return m_cDir; }
	int GetDirectoryEntryPos( CDirEntry* pDirEntry ) { return 0x0002 + (pDirEntry-m_cDir.m_acDirEntry) * 2; }
	int GetErrorID( void ) { return m_iIDError; }
	char* GetErrorStr( void ) { return m_szError; }
	BOOL GetUseExternalFlashStatus( void ) { return m_bUseExternalFlash; }
	int GetBlockSize( void ) { return m_iSizeBlock; }

	// operations
	void UseExternalFlash( BOOL bUseExternalFlash );
	BOOL CheckPresent( void );
	CDirEntry* FindFile( char* pszFile );
	BOOL Initialize( BOOL (*pfProgress)(int iPos, int iCount) );
	BOOL RemoveFile( char* pszFile );
	BOOL RemoveAllFiles( void );
	BOOL SetFileOrder( UINT* pauiPosOrder, UINT uiCount );
	BOOL TxDirectory( void );
	BOOL RxDirectory( void );
	BOOL TxFile( char* pszPathFile, BOOL (*pfProgress)(int iPos, int iCount) );
	BOOL RxFile( char* pszPathFile, BOOL (*pfProgress)(int iPos, int iCount) );
};

#endif		// _MPMF20_

